/*
Test code for a BCD programmed 858 PLL CB radio
Denys Parnell 20200911
At the moment, no changes are implimented for USB/LSB switching
Add a pair of 74HS595 shift registers.
from https://www.arduino.cc/en/Tutorial/ShftOut21
*/


#define Version "Cheetah_SSB_2_SR3"


#include "Wire.h"
#include "LiquidCrystal_PCF8574.h"


// synthesiser default settings on start up.
unsigned long frequency       = 28305000; //27005000;   // 4 bytes required for frequency
unsigned long MinFrequency    = 28305000; //27005000;   // minimum frequency
unsigned long MaxFrequency    = 29095000;   // maximum frequence

unsigned long step_size       = 10000;       // default tuning increment (Hz)
unsigned long IFoffset        = 7800000;    // IF offset

unsigned long AM_LSB          = 11284500;   //  
unsigned long USB             = 11285500;   // 
unsigned long PLLmixer        = (AM_LSB*3);   // 

int encoder;                                // encoder direction: CCW=-1; NC=0; CW=1

// encoder definitions
#define ENCODER_A  A0
#define ENCODER_B  A1

// A4 = SDA
// A5 = SCL

#define ENCODER_BUTTON 5                  // selects tuning decade

#define ENCODER_INPUT_PORT PINC
#define On  1
#define Off 0

//Pin connected to ST_CP of 74HC595
int latchPin = 8;
//Pin connected to SH_CP of 74HC595
int clockPin = 12;
////Pin connected to DS of 74HC595
int dataPin = 11;

LiquidCrystal_PCF8574 lcd(0x27); // set the LCD address to 0x27 for a 16 chars and 2 line display

void setup() {
  lcd.setBacklight(255);
//  lcd.begin(16, 2);                          // 16 char x 2 line LCD display
  lcd.begin(20, 4);                          // 16 char x 2 line LCD display
  lcd.print(Version);
  delay(2000);
  lcd.clear();
  // configure rotary encoder & push-button
  pinMode(ENCODER_A, INPUT_PULLUP);     // encoder leads
  pinMode(ENCODER_B, INPUT_PULLUP);
  pinMode(ENCODER_BUTTON, INPUT_PULLUP);

  pinMode(latchPin,OUTPUT);
  pinMode(clockPin,OUTPUT);
  pinMode(dataPin,OUTPUT);
  digitalWrite(latchPin,LOW);
  digitalWrite(clockPin,LOW);  
  digitalWrite(dataPin,LOW);
  
  encoder = read_encoder();                       // returns +1/0/-1 (cw/nc/ccw)
        displayFrequ();
        PLL();
}

/******************************
  main loop
*******************************/
void loop() {

  // check the frequency
  encoder = read_encoder();                       // returns +1/0/-1 (cw/nc/ccw)
  if (encoder != 0) {
    frequency = frequency + (encoder*step_size);
    frequency = constrain(frequency, MinFrequency, MaxFrequency);

    displayFrequ();
    PLL();
  }

}

/******************************
  read encoder direction
*******************************/
int read_encoder() {
  // definitions
#define DELAY 100                               // number of polling-loops   
  static unsigned short integrator;               // encoder debounce

  /* the array values are determined by combining the previous and current A1,A0
    Gray Code bit-patterns to form an array-index. +1 is assigned to all clockwise
    (CW) indexes; -1 is assigned to all counter-clockwise (CCW) indexes; and 0 is
    assigned to the remaining positions to indicate no movement*/
  static char encoder[] = {0, -1, 1, 0, 1, 0, 0, -1, -1, 0, 0, 1, 0, 1, -1, 0}; // SWAPPED PHASES
  //   static char encoder[] = {0,1,-1,0,-1,0,0,1,1,0,0,-1,0,-1,1,0};
  static unsigned char encoder_index = 0;

  static long count = 0;                         // CW rotation = +1; CCW rotation = -1
  static long this_reading;
  static long last_reading;
  char value;                                    // return value

  /* check for encoder rotation. The absolute "count" value changes as the encoder rotates.
    Typically there are four states (counts) per click, but often more due to contact bounce.
    Bounce doesn't matter so long as the "count" decreases" for CCW rotation; remains constant
    when there is no rotation; and increases with CW rotation. Bounce can be completely
    eliminated if we look for the "detent" code and integrate as shown below. */
  encoder_index <<= 2;                           // shift previous bit-pattern left two places
  encoder_index |= (ENCODER_INPUT_PORT & 0x03);  // get current bit-pattern and combine with previous
  count += encoder[( encoder_index & 0x0f )];

  // debounce encoder
  if ((ENCODER_INPUT_PORT & 0x03) == 3) {        // encoder on an indent (3 decimal == 00000011 binary)
    if (integrator < DELAY) {
      integrator++;
    }
  } else {                                       // encoder not resting on indent
    if (integrator > 0) {
      integrator--;
    }
  }

  // prepare the "return" value
  if (integrator >= DELAY) {                     // encoder deemed to be stationary
    this_reading = count;
    if (this_reading < last_reading) value = -1;
    if (this_reading == last_reading) value = 0;
    if (this_reading > last_reading) value = 1;
    last_reading = this_reading;                 // both readings now hold current "count"
    integrator = 0;                              // reset integrator
  } else {
    value = 0;
  }
  return (value);
}


void displayFrequ()   // Show the frequency
{
     long MHz = frequency/1000000;
  // convert frequency to floating point
  // adding 0.0000001 fixes a rounding error
    double freq = (double) frequency/1000000+0.0000001;
    lcd.setCursor(0,0);                          // start at char 0, line 2
    lcd.print("                    ");           // clear the line of previous displayed info
    lcd.setCursor(0,0);                          // start at char 2, line 1
    lcd.print("FRQ=");
    lcd.print(freq, 3);
    lcd.print(" MHz ");
}

void PLL()    // calculate the PLL programming inputs based on the frequency.
{
  int PLL_set = (((frequency+IFoffset)-PLLmixer)/step_size); // calculated here.
     lcd.setCursor(0,1);                          // start at char 0, line 2
     lcd.print("                    ");           // clear the line of previous displayed info
     lcd.setCursor(0,1);                          // start at char 0, line 2
     lcd.print("PLL=");
     lcd.print(PLL_set);                          // Display the PLL divisor 
     
     lcd.print(" BCD= ");
     int BCD;
     byte tempA=0;
     byte tempB=0;
     byte tempC=0;
     byte BCD_Low=0;
     byte BCD_High=0;

    tempC=PLL_set/100;                            // extract the decimal digits
    
    tempB=((PLL_set-(tempC*100))/10);
    
    tempA=(PLL_set-((tempC*100)+(tempB*10)));
   
     lcd.print(tempC);                           // display each decimal digit
     lcd.print(" ");
     lcd.print(tempB);
     lcd.print(" ");
     lcd.print(tempA);

    double freq = (double) (frequency+IFoffset)/1000000+0.0000001;
 
     lcd.setCursor(0,2);                          // start at char 0, line 3
     lcd.print("                    ");           // clear the line of previous displayed info
     lcd.setCursor(0,2);                          // start at char 0, line 3
     lcd.print("VFO=");
     lcd.print(freq, 3);                          // Display the VFO frequency
     lcd.print(" MHz ");

     lcd.setCursor(0,3);                          // start at char 0, line 4
     lcd.print("                    ");           // clear the line of previous displayed info
     lcd.setCursor(0,3);                          // start at char 0, line 4   


                                                  // rebuild the decimal digit  to the final hex number   
    BCD_Low = ((tempB << 4)| tempA);                                               
    BCD_High = tempC;
    
    lcd.print("High=");
    lcd.print ("0x");
    lcd.print(BCD_High,HEX);  
    lcd.print(" Low=");
    lcd.print ("0x");
    lcd.print(BCD_Low,HEX);  
    
    // display the final hex number that the 858 chip needs.



 
    
    digitalWrite(latchPin, 0);
    shiftOut(dataPin, clockPin, BCD_High);      // shift out high byte first
    shiftOut(dataPin, clockPin, BCD_Low);       // then the low byte
    digitalWrite(latchPin, 1);
    
}


void shiftOut(int myDataPin, int myClockPin, byte myDataOut) {
  // This shifts 8 bits out MSB first, 
  //on the rising edge of the clock,
  //clock idles low

//internal function setup
  int i=0;
  int pinState;
  pinMode(myClockPin, OUTPUT);
  pinMode(myDataPin, OUTPUT);

 //clear everything out just in case to
 //prepare shift register for bit shifting
  digitalWrite(myDataPin, 0);
  digitalWrite(myClockPin, 0);

  //for each bit in the byte myDataOut�

  for (i=7; i>=0; i--)  {
    digitalWrite(myClockPin, 0);

    if ( myDataOut & (1<<i) ) {
      pinState= 1;
    }
    else {  
      pinState= 0;
    }

    //Sets the pin to HIGH or LOW depending on pinState
    digitalWrite(myDataPin, pinState);
    //register shifts bits on upstroke of clock pin  
    digitalWrite(myClockPin, 1);
    //zero the data pin after shift to prevent bleed through
    digitalWrite(myDataPin, 0);
  }

  //stop shifting
  digitalWrite(myClockPin, 0);
}


 

